/*******************************************************************************
* YORKTOWN ADVANCED SGML PARSER (YASP)                                         *
*         An SGML Parser conforming to International Standard ISO 8879 --      *
*         Standard Generalized Markup Language                                 *
*------------------------------------------------------------------------------*
* Written 1987 - 1993 by                                                       *
*         Pierre G. Richard                                                    *
*         IBM T.J. Watson Research Center                                      *
*         Yorktown Heights, New York                                           *
*         petrus@watson.ibm.com                                                *
*------------------------------------------------------------------------------*
* FUNCTION NAME    = YASP / Document Instance Parsing                          *
* STATUS           = Version 0, Release 1, Level 36 and up                     *
* COMPONENT NAME   = teller.c                                                  *
* DESCRIPTIVE NAME = Interface to Application Notifications                    *
* DATE_WR          = 11/03/1988                                                *
* AUTHOR           = Pierre G. Richard - PETRUS at YKTVMX                      *
*                                                                              *
* FUNCTIONAL DESCRIPTION =                                                     *
*                                                                              *
* This module tells the Application about selected events.                     *
*                                                                              *
* CHANGE HISTORY =                                                             *
* Number Date     Inits    Change & reason (brief)                             *
*      1 01/15/89 PETRUS   Adapted to the new API conventions                  *
*      2 02/02/89 PETRUS   Reentrancy                                          *
*      3 07/10/89 PETRUS   Common Interface for all systems                    *
*      4 02/11/90 PETRUS   Field AHL replaced by (EXID *) in entity structure  *
*      5 03/01/91 DUROT    Add tll_SKIP and the POS structure                  *
*      6 08/28/91 PETRUS   DED storage now vanishes                            *
*      7 06/11/92 PETRUS   ETY_PFC can return an ETY                           *
*      8 04/01/97 ESPERT   New event reporting for markup declaration          *
*******************************************************************************/

/*-------------+
|Include Files |
+-------------*/
#include "banner.h"    /* This module includes code depending on YASP version */
#include "machine.h"
#include "portable.h"
#include "yaspapi.h"
#include "global.h"
#include "usrlib.h"
#include "sgml.h"
#include "modele.h"
#include "teller.h"

/*--------------+
|Local Functions|
+--------------*/
static void LP_FCT tell_ooc_position(GA_P);

/*-STATIC----------------------------------------------------tell_ooc_position-+
|Tell the position of Out Of Context Data                                      |
+-----------------------------------------------------------------------------*/
static void LP_FCT tell_ooc_position(GA_P gbla_p)
{
   PAC *pac_p;
   int rc;

   if (GA(ooc_position).ety_p != NULL) {         /* only if we have to */
      pac_p = gbla_p->pac_p;
      pac_p->pos = GA(ooc_position);
      pac_p->sdc = ERROR_SDC;
      pac_p->func = SKIP_PFC;
      if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
      GA(ooc_position).ety_p=NULL;               /* No more OOCDATA */
   }
}

/*--------------------------------------------------------------------tll_init-+
|This is entered when the Parser starts                                        |
+-----------------------------------------------------------------------------*/
void P_FCT tll_init(void)
{
}

/*---------------------------------------------------------------------tll_STG-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void P_FCT tll_STG(GA_P gbla_p, DED *ded_p)
{
   int rc;
   register PAC *pac_p = gbla_p->pac_p;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);                /* Out Of Context Data */
#endif
   ++(GA(curr_elt_id));
   pac_p->dad._.p = (uchar *)(ded_p);
#if YASP_RVL_GE(0,1,26)
   /* Do NOT update position when implied: keep previously reported pos. */
   if (!(ded_p->is & IMP_LF)) em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ELTST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*---------------------------------------------------------------------tll_ETG-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ETG(GA_P gbla_p, DED *ded_p)
{
   int rc;
   register PAC *pac_p = gbla_p->pac_p;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);                /* Out Of Context Data */
#endif
   pac_p->dad._.p = (uchar *)(ded_p);
#if YASP_RVL_GE(0,1,26)
   /* Do NOT update position when implied: keep previously reported pos. */
   if (!(ded_p->is & IMP_LF) && (ded_p->tagmin != DATA_ET_MNF)) {
      em_pos(gbla_p, &(pac_p->pos), 1);
   }
#endif
   pac_p->func = ELTND_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_DATA-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void P_FCT tll_DATA(GA_P gbla_p, uchar *buf, uint length, POS *pos_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   pac_p->pos = *pos_p;
#endif
   pac_p->dad.len = length;
   pac_p->dad._.p = buf;
   pac_p->func = TEXT_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*----------------------------------------------------------------------tll_RE-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void P_FCT tll_RE(GA_P gbla_p, POS *pos_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   pac_p->pos = *pos_p;
#endif
   pac_p->func = RE_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*----------------------------------------------------------------------tll_PI-+
| Notify a processing instruction                                              |
+-----------------------------------------------------------------------------*/
void P_FCT tll_PI(GA_P gbla_p, uchar *pi_lit, uint pi_len)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->dad.len = pi_len;
   pac_p->dad._.p = pi_lit;
   pac_p->func = PI_PFC;
   rc = pac_p->ase(pac_p);
   switch (rc) {
   default:
      exit(rc);
   case 0:
      break;
   }
}

/*---------------------------------------------------------------------tll_ePI-+
| Notify a processing instruction from a PI entity                             |
| As tell_PI, except this notification is encapsulated by the ETYST/ETYCL.     |
+-----------------------------------------------------------------------------*/
ETY * P_FCT tll_ePI(GA_P gbla_p, ETY *ety_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   struct {                              /* Dummy YB, to be consistent */
      OEB  oeb;
      OEB *prev_oeb_p;
   }dummy_yb;
   int rc;
   ETY *ret_ety_p = (ETY *)0;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);            /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 1);     /* Position within the parent entity */
#endif
   dummy_yb.oeb.ety_p = ety_p;
   dummy_yb.oeb.offset = dummy_yb.oeb.col = 1;
   dummy_yb.oeb.row = 0;
   dummy_yb.oeb.achl.p = (void *)0;
   dummy_yb.oeb.appl_info= (uchar *)0;
   dummy_yb.prev_oeb_p = pac_p->oeb_p;
   pac_p->oeb_p = &dummy_yb.oeb;
   pac_p->dad._.p = ety_p;
   pac_p->func = ETYST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);

   pac_p->dad.len = ety_p->val.lit.len;
   pac_p->dad._.p = ety_p->val.lit._.p;
   pac_p->func = PI_PFC;
   rc = pac_p->ase(pac_p);
   if (rc != 0) exit(rc);

   pac_p->oeb_p = dummy_yb.prev_oeb_p;
   pac_p->dad._.p = ety_p;
   pac_p->func = ETYND_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
   return(ret_ety_p);
}

/*---------------------------------------------------------------------tll_ETY-+
|   ETY_PFC return codes:                                                      |
|                                                                              |
|   RC  2:  Application wants to include an ETY at this point                  |
|   RC  4:  Application does not recognize the given SDATA                     |
|           (Application has issued its own moangram)                          |
|   RC  8:  Unrecognized data entity type  (cannot occur !)                    |
+-----------------------------------------------------------------------------*/
ETY * P_FCT tll_ETY(GA_P gbla_p, ETY *ety_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->dad._.p = (uchar *)ety_p;
   pac_p->func = ETY_PFC;
   rc = pac_p->ase(pac_p);
   switch (rc) {
   default:
      exit(rc);
   case 0:
      break;
   case 1:
      return((ETY *)pac_p->dad._.p);
   }
   return((ETY *)0);
}

/*---------------------------------------------------------------------tll_MAP-+
| Tell a USEMAP was encountered in document instance                           |
+-----------------------------------------------------------------------------*/
void P_FCT tll_MAP(GA_P gbla_p, SRM *srm_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->dad._.p = srm_p;
   pac_p->func = MAP_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*----------------------------------------------------------------------tll_SR-+
| Tell a SHORTREF mapped to an existing entity was found                       |
+-----------------------------------------------------------------------------*/
void P_FCT tll_SR(GA_P gbla_p, SRM *srm_p, int sr_id)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->dad._.p = srm_p;
   pac_p->dad.len = sr_id;
   pac_p->func = SR_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_NGML-+
| Tell a non-SGML character has been found                                     |
+-----------------------------------------------------------------------------*/
void P_FCT tll_NGML(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   tell_ooc_position(gbla_p);             /* Out Of Context Data */
   em_pos(gbla_p, &(pac_p->pos), 0);
#endif
   pac_p->dad._.p = (uchar *)&(GA(curr_char));
   pac_p->dad.len = 1;
   pac_p->func = NSGML_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

#if YASP_RVL_GE(0,1,26)
/*--------------------------------------------------------------------tll_SKIP-+
| Tell some skipped data                                                       |
+-----------------------------------------------------------------------------*/
void P_FCT tll_SKIP(GA_P gbla_p, SDC sdc)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

   tell_ooc_position(gbla_p);             /* Out Of Context Data if any */
   em_pos(gbla_p, &(pac_p->pos), 1);
   pac_p->sdc = sdc;
   pac_p->func = SKIP_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}
#endif

/*--------------------------------------------------------------------tll_DECB-+
| Tell we start parsing the SGML declaration                                   |
+-----------------------------------------------------------------------------*/
void P_FCT tll_DECB(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

   pac_p->func = DECST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_DTDB-+
| Tell we start parsing a DTD                                                  |
+-----------------------------------------------------------------------------*/
void P_FCT tll_DTDB(GA_P gbla_p, POS *pos_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   pac_p->pos = *pos_p;
#endif
   pac_p->func = DTDST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_DTDN-+
| Tell we ended parsing a DTD                                                  |
+-----------------------------------------------------------------------------*/
void P_FCT tll_DTDN(GA_P gbla_p, uchar *dtdname_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->dad.len = strlen( pac_p->dad._.p = dtdname_p );
   pac_p->func = DTDND_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_PROB-+
| Tell we start parsing the Prolog                                             |
+-----------------------------------------------------------------------------*/
void P_FCT tll_PROB(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = PROST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_PRON-+
| Tell the end of Prolog has been reached                                      |
+-----------------------------------------------------------------------------*/
void P_FCT tll_PRON(GA_P gbla_p, ELT *elt_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = PROND_PFC;
   pac_p->dad._.p = (uchar *)elt_p;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ETDCST-+
| Tell the start of an entity declaration has been reached                     |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ETDCST(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ETYDCLST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ETDCND-+
| Tell the end of an entity declaration has been reached                       |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ETDCND(GA_P gbla_p, ETY *ety_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ETYDCLND_PFC;
   pac_p->dad._.p = (uchar *)(ety_p);
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_NTDCST-+
| Tell the start of a notation declaration has been reached                    |
+-----------------------------------------------------------------------------*/
void P_FCT tll_NTDCST(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = NOTDCLST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_NTDCND-+
| Tell the end of a notation declaration has been reached                      |
+-----------------------------------------------------------------------------*/
void P_FCT tll_NTDCND(GA_P gbla_p, NCB *ncb_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = NOTDCLND_PFC;
   pac_p->dad._.p = (uchar *)(ncb_p);
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ELDCST-+
| Tell the start of an element declaration has been reached                    |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ELDCST(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ELTDCLST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ELDCND-+
| Tell the end of an element declaration has been reached                      |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ELDCND(GA_P gbla_p, ELT *elt_p, uint elt_count)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ELTDCLND_PFC;
   pac_p->dad._.p = (uchar *)(elt_p);
   pac_p->dad.len = elt_count;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_AEDCST-+
| Tell the start of an attlist declaration for element(s) has been reached     |
+-----------------------------------------------------------------------------*/
void P_FCT tll_AEDCST(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ATLELTDCLST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_AEDCND-+
| Tell the end of an attlist declaration for element(s) has been reached       |
+-----------------------------------------------------------------------------*/
void P_FCT tll_AEDCND(GA_P gbla_p, ELT *elt_p, uint elt_count)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ATLELTDCLND_PFC;
   pac_p->dad._.p = (uchar *)(elt_p);
   pac_p->dad.len = elt_count;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ANDCST-+
| Tell the start of an attlist declaration for notation(s) has been reached    |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ANDCST(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ATLNOTDCLST_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*------------------------------------------------------------------tll_ANDCND-+
| Tell the end of an attlist declaration for notation(s) has been reached      |
+-----------------------------------------------------------------------------*/
void P_FCT tll_ANDCND(GA_P gbla_p, NCB *ncb_p, uint ncb_count)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

#if YASP_RVL_GE(0,1,26)
   em_pos(gbla_p, &(pac_p->pos), 1);
#endif
   pac_p->func = ATLNOTDCLND_PFC;
   pac_p->dad._.p = (uchar *)(ncb_p);
   pac_p->dad.len = ncb_count;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}

/*--------------------------------------------------------------------tll_DOCN-+
| Tell the End of DOC has been reached                                         |
+-----------------------------------------------------------------------------*/
void P_FCT tll_DOCN(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

   pac_p->func = DOCND_PFC;
   if ((rc=pac_p->ase(pac_p))!=0) exit(rc);
}


/*--------------------------------------------------------------------tll_DAPN-+
| Tell the End of DOC right after Prolog                                       |
+-----------------------------------------------------------------------------*/
boolean P_FCT tll_DAPN(GA_P gbla_p)
{
   register PAC *pac_p = gbla_p->pac_p;
   int rc;

   pac_p->func = DAPND_PFC;
   rc = pac_p->ase(pac_p);
   switch (rc) {
   case 0:
      return(FALSE);
   case 4321:
      /*-----------------------------------------------------+
      | This is an undocumented return code from DAPND_PFC.  |
      | Its purpose is to parse a DTD only and return to     |
      | the Application while the YASP storage for the DTD   |
      | is still available.  The Application can therefore   |
      | uses specific YASP services to analyse the DTD.      |
      | Warning: do not use this for a SUBDOC!               |
      +-----------------------------------------------------*/
      return(TRUE);
   default:
      exit(rc);
   }
}
/******************************************************************************/
